/*
 * Copyright (C) 2019 by Bundesamt für Justiz, Fachstelle für Rechtsinformatik
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
package ch.admin.suis.msghandler.common;

import java.io.ByteArrayInputStream;
import java.nio.charset.StandardCharsets;
import java.util.AbstractMap.SimpleEntry;
import org.apache.commons.digester3.Digester;
import org.apache.xerces.parsers.SAXParser;
import static org.junit.Assert.*;
import org.junit.Test;
import org.xml.sax.InputSource;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;

/**
 *
 * @author Adrian Greiler
 */
public class XmlParserConfiguratorTest
{

  /*
   * see https://github.com/OWASP/CheatSheetSeries/blob/master/cheatsheets/XML_External_Entity_Prevention_Cheat_Sheet.md
   */

  private static SimpleEntry[] SAXPARSER_FEATURES = new SimpleEntry[]{
      new SimpleEntry<String, Boolean>("http://apache.org/xml/features/disallow-doctype-decl", true),
      new SimpleEntry<String, Boolean>("http://xml.org/sax/features/external-general-entities", false),
      new SimpleEntry<String, Boolean>("http://xml.org/sax/features/external-parameter-entities", false),
      new SimpleEntry<String, Boolean>("http://apache.org/xml/features/nonvalidating/load-external-dtd", false)};
  
  
  public XmlParserConfiguratorTest()
  {
  }

  /**
   * Test of hardenDigesterAgainstXXE method, of class XmlParserConfigurator.
   */
  @Test
  public void testHardenDigesterAgainstXXE_SAXParser_PlainFeaturesetTest() throws Exception
  {
    SAXParser parser = new SAXParser();
    XmlParserConfigurator.hardenDigesterAgainstXXE(parser);

    for (SimpleEntry<String, Boolean> p : SAXPARSER_FEATURES)
    {
      assertEquals(p.getValue(), parser.getFeature(p.getKey()));
    }

    // ensure, that we do not use xercers
    try
    {
      parser.getFeature("http://xerces.apache.org/xerces-j/features.html#external-general-entities");
      fail("we are using Xerces -> OWASP best practices need to be applied accordingly!");
    } catch(SAXNotRecognizedException | SAXNotSupportedException ex)
    {
      // ok
    }

    // nor xerces 2
    try
    {
      parser.getFeature("http://xerces.apache.org/xerces2-j/features.html#external-general-entities");
      fail("we are using Xerces 2 -> OWASP best practices need to be applied accordingly!");
    } catch(SAXNotRecognizedException | SAXNotSupportedException ex)
    {
      // ok
    }
  }

  // org.xml.sax.SAXParseException; lineNumber: 2; columnNumber: 10;
  // DOCTYPE is disallowed when the feature "http://apache.org/xml/features/disallow-doctype-decl" set to true.
  @Test(expected = org.xml.sax.SAXParseException.class)
  public void testHardenDigesterAgainstXXE_SAXParser_XxeTest() throws Exception
  {
    SAXParser parser = new SAXParser();
    XmlParserConfigurator.hardenDigesterAgainstXXE(parser);

    String xxeDtd = "target/test-classes/common/XmlParserConfiguratorTest.dtd";

    parser.parse(new InputSource(new ByteArrayInputStream(
        ("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
        "<!DOCTYPE xxe SYSTEM \"" + xxeDtd + "\">\n" +
        "<blub>fail</blub>").getBytes(StandardCharsets.UTF_8))));
    
    fail("An external dtd should not be loaded: "+xxeDtd);
  }

  /**
   * Test of hardenDigesterAgainstXXE method, of class XmlParserConfigurator.
   */
  @Test
  public void testHardenDigesterAgainstXXE_Digester_PlainFeaturesetTest() throws Exception
  {
    Digester digester = new Digester();
    XmlParserConfigurator.hardenDigesterAgainstXXE(digester);

    for (SimpleEntry<String, Boolean> p : SAXPARSER_FEATURES)
    {
      assertEquals(p.getValue(), digester.getFeature(p.getKey()));
    }

    assertFalse(digester.getXIncludeAware());

    // ensure, that we do not use xercers
    try
    {
      digester.getFeature("http://xerces.apache.org/xerces-j/features.html#external-general-entities");
      fail("we are using Xerces -> OWASP best practices need to be applied accordingly!");
    } catch(SAXNotRecognizedException | SAXNotSupportedException ex)
    {
      // ok
    }

    // nor xerces 2
    try
    {
      digester.getFeature("http://xerces.apache.org/xerces2-j/features.html#external-general-entities");
      fail("we are using Xerces 2 -> OWASP best practices need to be applied accordingly!");
    } catch(SAXNotRecognizedException | SAXNotSupportedException ex)
    {
      // ok
    }
  }

  
  // org.xml.sax.SAXParseException; lineNumber: 2; columnNumber: 10;
  // DOCTYPE is disallowed when the feature "http://apache.org/xml/features/disallow-doctype-decl" set to true.
  @Test(expected = org.xml.sax.SAXParseException.class)
  public void testHardenDigesterAgainstXXE_Digester_XxeTest() throws Exception
  {
    Digester digester = new Digester();
    XmlParserConfigurator.hardenDigesterAgainstXXE(digester);

    String xxeDtd = "target/test-classes/common/XmlParserConfiguratorTest.dtd";

    digester.parse(new InputSource(new ByteArrayInputStream(
        ("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
        "<!DOCTYPE xxe SYSTEM \"" + xxeDtd + "\">\n" +
        "<blub>fail</blub>").getBytes(StandardCharsets.UTF_8))));
    
    fail("An external dtd should not be loaded: "+xxeDtd);
  }
}